/*
// THIS FILE IS COPYRIGHT 1994 BY TELENOR RESEARCH, NORWAY
//
//                    ALL RIGHTS RESERVED.
//
// For rights granted to the public see the COPYRIGHT file
*/

/* Programming : Jon Stephenson von Tetzchner		 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "trans.h"

#define debug(x) if (DEBUG) {fprintf(stderr, x);}

#define FALSE 0     		/* Various useful constants */
#define TRUE 1
#define True 1
#define False 0

#define UNKNOWN 	0	/* Paragraph types */

#define HEADING0 	1	/* Headings */
#define HEADING1 	2
#define HEADING2 	3
#define HEADING3 	4
#define HEADING4 	5
#define HEADING5 	6
#define HEADING6 	7

#define PROGRAM  	9
#define REFERENCE 	10
#define TITLE           11
#define TOC		12
#define ADDRESS		13
#define AUTHOR	 	14
#define ABSTRACT 	15
#define LOCALTITLE	16
#define BODY     	17

#define COMREFERENCE	18
#define ACRONYM		20 

#define TABLE	    	23

#define FIGURE		25

#define LISTINDENT1	26	/* Non-consecutive lists with full indent */
#define LISTINDENT2	27	/* (HTML: <DL><DD>item) */
#define LISTINDENT3	28
#define LISTINDENT4	29
#define LISTINDENT5	30
#define LISTINDENT6	31

#define LISTBULLET1	32	/* Non-consecutive lists with bullet as mark */
#define LISTBULLET2	33	/* (HTML: <UL><LI>item) */
#define LISTBULLET3	34
#define LISTBULLET4	35
#define LISTBULLET5	36
#define LISTBULLET6	37

#define LISTMARK1 	38	/* Lists using MIF PgfNumString as mark */
#define LISTMARK2 	39	/* (HTML: <DL><DT>mark<DD>item) */
#define LISTMARK3 	40
#define LISTMARK4 	41
#define LISTMARK5 	42
#define LISTMARK6 	44

#define LISTNUMBER1 	45	/* Numbered consecutive lists */
#define LISTNUMBER2 	46	/* (HTML: <OL><LI>item) */
#define LISTNUMBER3 	47
#define LISTNUMBER4 	48
#define LISTNUMBER5 	49
#define LISTNUMBER6 	50

#define LISTCONTINUE1 	51	/* Additional paragraphs in any type of list */
#define LISTCONTINUE2 	52
#define LISTCONTINUE3 	53
#define LISTCONTINUE4 	54
#define LISTCONTINUE5 	55
#define LISTCONTINUE6 	56

#define CELLHEADING	57
#define CELLBODY	58

#define UHEADING0 	59	/* Headings not shown in table of contents */
#define UHEADING1 	60
#define UHEADING2 	61
#define UHEADING3 	62
#define UHEADING4 	63
#define UHEADING5 	64
#define UHEADING6 	65

#define INDENTHEADING0  69	/* Indented headings */
#define INDENTHEADING1	70
#define INDENTHEADING2	71
#define INDENTHEADING3	72
#define INDENTHEADING4	73
#define INDENTHEADING5	74
#define INDENTHEADING6	75

#define GLOSSARY1   	80
#define GLOSSARY2	81
#define GLOSSARY3	82
#define COMGLOSSARY1	83
#define COMGLOSSARY2	84
#define COMGLOSSARY3	85

#define HTML		100	/* Direct HTML code */

#define UNORDERED	0	/* HTML list types */
#define ORDERED		1
#define DESCRIPTION	2

#define MAXDEPTH	6	/* Maximum list levels */
#define STACKSIZE	(MAXDEPTH + 1)

int depth = 0;		/* Current level of list nesting */
int tcdepth = 0;	/* Current level of table of contents nesting */
int stack[STACKSIZE];	/* Type of list found at each level of nesting */

#define MAXNUMFOOTNOTES 	100		
#define MAXNUMREFERENCES 	500

#ifndef kEM_DASH_PREF
#define kEM_DASH_PREF "---"
#endif

#ifndef kEN_DASH_PREF
#define kEN_DASH_PREF "--"
#endif

struct FMPgf *pgf_list = NULL;
struct FMPgf *current_pgf = NULL;

struct FMFont *font_list = NULL;
struct FMFont *current_font = NULL;

struct FMImportObject *impobj_list = NULL;
struct FMImportObject *current_impobj = NULL;

int in_pgf_catalog = FALSE;	/* Indicates when we're inside the paragraph
				   catalog.  Thus format information can be
				   extracted from paragraphs in the catalog. */

int in_font_catalog = FALSE;	/* Indicates when we're in the font
				   catalog.  Thus format information can be
				   extracted from fonts in the catalog. */

int in_body_page = FALSE;	/* Indicates when we're inside a body
				   page.  Thus any ImportObjects in the page
				   can be extracted. */

int in_string = FALSE;		/* Indicates when we're inside a string.
				   Thus any following strings can be added to
				   the current paragraph. */

int in_font = FALSE;
int in_textflow = FALSE;
int in_reference = FALSE;
int in_glossary = FALSE;
int in_program = FALSE;
int in_acronym = FALSE;
int in_tables = FALSE;
int in_footnote = FALSE;		
int in_frame = FALSE;
int in_inline_frame = FALSE;

int in_heading = FALSE;		/* This refers to TITLE and LOCALTITLE
				   as well as headings, since they are
				   translated to <H1> */

int in_code = FALSE;		/* Whether between <CODE> and </CODE> */

#define NORMAL 0
#define SYMBOL 1
#define DINGBATS 2
#define COURIER 3

int font_type = NORMAL;

int font_bold = FALSE;			/* Whether font currently bold */
int font_italic = FALSE;		/* Whether font currently italic */
int font_underline = FALSE;		/* Whether font currently underline */
int font_subscript = FALSE;		/* Whether font currently subscript */
int font_superscript = FALSE;		/* Whether font currently superscript*/

char font_openers[50];			/* When a font is encountered, opening
					   tags are stored here temporarily so
					   all closing tags are inserted into
					   output before any opening ones (e.g.
					   </B> before <I>).  This is because
					   some browsers don't like
					   overlapping tags. */

int marker_number;			/* Paragraph marker ID */
int marker_type;			/* Marker type */
char *marker_text = NULL;		/* Marker text for hypertext markers */
int non_finished_marker = FALSE;

int header = -1;			/* Identifies paragraph type using */
                                        /* the types specified above */

int header_id = 0;			/* Used to discriminate header ids */

char *footnote[MAXNUMFOOTNOTES];	/* Holds all footnotes found which
					   are written in the end */
int footnote_use_count = 0;		
int footnote_def_count = 0;		/* Assume that footnotes come one
					   after one in use and def */

int table_num = 1;		/* Faar aa gjenkjenne figurer */
int figur_num = 1;
int inlineframe_num = 0;	/* Number of inline frames */

char *reference_filename;	/* Holds current reference filename */
int reference_number;		/* Holds reference id of current reference */

int separator_pending = FALSE;	/* Whether separator should be inserted after
				   current paragraph */

char *para_num_string;		/* Holds current paragraph numbering */
char *para_num_font;		/* Holds current paragraph numbering font */

char *html_title_string = NULL;	/* HREF and title to insert into table of
				   contents */
int title_printed = FALSE;	/* Whether html_title_string written to 
				   table of contents yet */

char *text_flow_tag;		/* Holds tag of current text flow */

char *TRANSFilename;	       	/* Name of file to be created by trans
				   shell script */
char *TRANSBasename;       	/* Basename of TRANSFilename (path stripped) */
char *TITLEFilename;		/* File for title and other HTML header info
				   (TRANSFilename.tmp.000T) */
char *INDEXFilename;		/* File for generated table of contents
				   (TRANSFilename.tmp.000I) */
char *HEADFilename;		/* File for header portion of input file,
				   which is prepended to all figure and table
				   files (TRANSFilename.tmp.000H) */
char *TAGSFilename;		/* File holding fm-tags to html-tags 
				   conversion info */

char pagefilename[300];
char flowfilename[300];

FILE *pagefile;			/* File to hold page info */
FILE *flowfile;			/* File to hold textflow */
FILE *f;			/* Points to one of the two above */
FILE *indexfile;		/* File to hold index */
FILE *titlefile;		/* File to hold title */
FILE *tagsfile;

char fm_tags[500][100];
int int_tags[500];
int tags_count = 0;

char html_tags[100][100];
int  htno = 0;

char *unknown_int_tags[200];
int unknown_int_tags_count = 0;
char *unknown_ext_tags[200];
int unknown_ext_tags_count = 0;

int tabfig[500];		/* Holds indicators of TABLEs and FIGUREs */
int tabfig_ids[500];		
int tabfig_count = 0;

char paragraph[50000];		/* Holds the current paragraph */

int textrect_ids[100];		/* ID's of all textrects used inside the
				   current frame */
int textrect_count = 0;		/* The number of textrects inside the current
				   frame */
int current_textrect_id = -1;

int frame_ids[10000];		/* Keeping track of which frame (the value)
				   contains given textrect (the index) */

int current_frame_id = -1;
FILE *framef;			/* Output file for current figure */
char frame_filename[300];
FILE *frametmpf;		/* Temporary file for current figure */
char frame_tmpname[300];
char *frame_filenames[1000];

int frame_count = 0;		/* total number of frames */
int aframe_count = 0;		/* number of frames inside AFrames section */
int frame_id_type[1000][2];	/* frame_id_type[n][0] = Frame id number */
                                /* frame_id_type[n][1] = Frame type, either
				   INLINEFRAME or OTHERFRAME */
#define INLINEFRAME	0
#define OTHERFRAME  	1

char sys_call[10000];


char *XDPI;			/* Horizontal graphics resolution
				   in dots per inch */
char *YDPI;			/* Vertical graphics resolution
				   in dots per inch */
char *NOCROP;			/* Whether to omit cropping of 
				   graphics */
char *NOGRAPHICS;		/* Whether to omit conversion of
				   contents of frames */
char *NOTOC;			/* Whether to omit generation of a
				   table of contents */
char *STRIP;			/* Whether ".doc" extension has been
				   stripped from document files */
char *QUIET;			/* Whether to print error messages
				   only, not status messages */
char *DEBUG;			/* Whether to print debug information
				   messages */


/*****************************************************************************
  int GetIntTag() :
 *****************************************************************************/

int GetIntTag( tag )
      char* tag;
{
    if (strcmp(tag,"UNKNOWN") == 0) return UNKNOWN;
    else if (strcmp(tag,"HEADING0") == 0) return HEADING0;
    else if (strcmp(tag,"HEADING1") == 0) return HEADING1;
    else if (strcmp(tag,"HEADING2") == 0) return HEADING2;
    else if (strcmp(tag,"HEADING3") == 0) return HEADING3;
    else if (strcmp(tag,"HEADING4") == 0) return HEADING4;
    else if (strcmp(tag,"HEADING5") == 0) return HEADING5;
    else if (strcmp(tag,"HEADING6") == 0) return HEADING6;
    else if (strcmp(tag,"UHEADING0") == 0) return UHEADING0;
    else if (strcmp(tag,"UHEADING1") == 0) return UHEADING1;
    else if (strcmp(tag,"UHEADING2") == 0) return UHEADING2;
    else if (strcmp(tag,"UHEADING3") == 0) return UHEADING3;
    else if (strcmp(tag,"UHEADING4") == 0) return UHEADING4;
    else if (strcmp(tag,"UHEADING5") == 0) return UHEADING5;
    else if (strcmp(tag,"UHEADING6") == 0) return UHEADING6;
    else if (strcmp(tag,"INDENTHEADING0") == 0) return INDENTHEADING0;
    else if (strcmp(tag,"INDENTHEADING1") == 0) return INDENTHEADING1;
    else if (strcmp(tag,"INDENTHEADING2") == 0) return INDENTHEADING2;
    else if (strcmp(tag,"INDENTHEADING3") == 0) return INDENTHEADING3;
    else if (strcmp(tag,"INDENTHEADING4") == 0) return INDENTHEADING4;
    else if (strcmp(tag,"INDENTHEADING5") == 0) return INDENTHEADING5;
    else if (strcmp(tag,"INDENTHEADING6") == 0) return INDENTHEADING6;
    else if (strcmp(tag,"BODY") == 0) return BODY;
    else if (strcmp(tag,"PROGRAM") == 0) return PROGRAM;
    else if (strcmp(tag,"REFERENCE") == 0) return REFERENCE;
    else if (strcmp(tag,"COMREFERENCE") == 0) return COMREFERENCE;
    else if (strcmp(tag,"TITLE") == 0) return TITLE;
    else if (strcmp(tag,"LOCALTITLE") == 0) return LOCALTITLE;
    else if (strcmp(tag,"TOC") == 0) return TOC;
    else if (strcmp(tag,"ADDRESS") == 0) return ADDRESS;
    else if (strcmp(tag,"AUTHOR") == 0) return AUTHOR;
    else if (strcmp(tag,"ABSTRACT") == 0) return ABSTRACT;
    else if (strcmp(tag,"TABLE") == 0) return TABLE;
    else if (strcmp(tag,"FIGURE") == 0) return FIGURE;
    else if (strcmp(tag,"GLOSSARY1") == 0) return GLOSSARY1;
    else if (strcmp(tag,"GLOSSARY2") == 0) return GLOSSARY2;
    else if (strcmp(tag,"GLOSSARY3") == 0) return GLOSSARY3;
    else if (strcmp(tag,"COMGLOSSARY1") == 0) return COMGLOSSARY1;
    else if (strcmp(tag,"COMGLOSSARY2") == 0) return COMGLOSSARY2;
    else if (strcmp(tag,"COMGLOSSARY3") == 0) return COMGLOSSARY3;
    else if (strcmp(tag,"ACRONYM") == 0) return ACRONYM;
    else if (strcmp(tag,"LISTINDENT1") == 0) return LISTINDENT1;
    else if (strcmp(tag,"LISTINDENT2") == 0) return LISTINDENT2;
    else if (strcmp(tag,"LISTINDENT3") == 0) return LISTINDENT3;
    else if (strcmp(tag,"LISTINDENT4") == 0) return LISTINDENT4;
    else if (strcmp(tag,"LISTINDENT5") == 0) return LISTINDENT5;
    else if (strcmp(tag,"LISTINDENT6") == 0) return LISTINDENT6;
    else if (strcmp(tag,"LISTBULLET1") == 0) return LISTBULLET1;
    else if (strcmp(tag,"LISTBULLET2") == 0) return LISTBULLET2;
    else if (strcmp(tag,"LISTBULLET3") == 0) return LISTBULLET3;
    else if (strcmp(tag,"LISTBULLET4") == 0) return LISTBULLET4;
    else if (strcmp(tag,"LISTBULLET5") == 0) return LISTBULLET5;
    else if (strcmp(tag,"LISTBULLET6") == 0) return LISTBULLET6;
    else if (strcmp(tag,"LISTMARK1") == 0) return LISTMARK1;
    else if (strcmp(tag,"LISTMARK2") == 0) return LISTMARK2;
    else if (strcmp(tag,"LISTMARK3") == 0) return LISTMARK3;
    else if (strcmp(tag,"LISTMARK4") == 0) return LISTMARK4;
    else if (strcmp(tag,"LISTMARK5") == 0) return LISTMARK5;
    else if (strcmp(tag,"LISTMARK6") == 0) return LISTMARK6;
    else if (strcmp(tag,"LISTNUMBER1") == 0) return LISTNUMBER1;
    else if (strcmp(tag,"LISTNUMBER2") == 0) return LISTNUMBER2;
    else if (strcmp(tag,"LISTNUMBER3") == 0) return LISTNUMBER3;
    else if (strcmp(tag,"LISTNUMBER4") == 0) return LISTNUMBER4;
    else if (strcmp(tag,"LISTNUMBER5") == 0) return LISTNUMBER5;
    else if (strcmp(tag,"LISTNUMBER6") == 0) return LISTNUMBER6;
    else if (strcmp(tag,"LISTCONTINUE1") == 0) return LISTCONTINUE1;
    else if (strcmp(tag,"LISTCONTINUE2") == 0) return LISTCONTINUE2;
    else if (strcmp(tag,"LISTCONTINUE3") == 0) return LISTCONTINUE3;
    else if (strcmp(tag,"LISTCONTINUE4") == 0) return LISTCONTINUE4;
    else if (strcmp(tag,"LISTCONTINUE5") == 0) return LISTCONTINUE5;
    else if (strcmp(tag,"LISTCONTINUE6") == 0) return LISTCONTINUE6;
    else if (strcmp(tag,"CELLHEADING") == 0) return CELLHEADING;
    else if (strcmp(tag,"CELLBODY") == 0) return CELLBODY;
    else if (strncmp(tag, "HTML",4) == 0 )
	{
	    /* mnh: HTML tags implemented as HTML beginning-tag ; ending-tag */
	    tag += 4;
	    while (*tag && (*tag == ' ' || *tag == '\t'))
	        tag++;
	    strcpy(html_tags[htno], tag);
	    return HTML + htno++;
	}
    else 
	{
	    int i;
	    for (i = 0; i < unknown_int_tags_count; i++)
		if (strcmp(tag, unknown_int_tags[i]) == 0)
		    break;

	    if (i == unknown_int_tags_count)
		{
		    unknown_int_tags[unknown_int_tags_count++] = strdup(tag);
		    fprintf(stderr, "trans: Unknown internal tag: \"%s\"\n",
			    tag);
		}

	    return UNKNOWN;
	}
}



/*****************************************************************************
 int IsIndent()
 *****************************************************************************/

int IsIndent(tag)
      int tag;
{
    return tag >= LISTINDENT1 && tag <= LISTINDENT6 ? TRUE : FALSE;
}


/*****************************************************************************
 int IsBullet()
 *****************************************************************************/

int IsBullet(tag)
      int tag;
{
    return tag >= LISTBULLET1 && tag <= LISTBULLET6 ? TRUE : FALSE;
}


/*****************************************************************************
 int IsMark()
 *****************************************************************************/

int IsMark(tag)
      int tag;
{
    return tag >= LISTMARK1 && tag <= LISTMARK6 ? TRUE : FALSE;
}


/*****************************************************************************
 int IsNumber()
 *****************************************************************************/

int IsNumber(tag)
      int tag;
{
    return tag >= LISTNUMBER1 && tag <= LISTNUMBER6 ? TRUE : FALSE;
}


/*****************************************************************************
 int IsList()
 *****************************************************************************/

int IsList(tag)
      int tag;
{
    return IsIndent(tag) || IsBullet(tag) || IsMark(tag) || IsNumber(tag);
}


/*****************************************************************************
 int IsContinue()
 *****************************************************************************/

int IsContinue(tag)
      int tag;
{
    return tag >= LISTCONTINUE1 && tag <= LISTCONTINUE6 ? TRUE : FALSE;
}


/*****************************************************************************
 int Level() : return the nesting level associated with the given bullet or
	       list tag
 *****************************************************************************/

int Level(tag)
      int tag;
{
    if (IsIndent(tag))
	return tag + 1 - LISTINDENT1;
    else if (IsBullet(tag))
	return tag + 1 - LISTBULLET1;
    else if (IsMark(tag))
	return tag + 1 - LISTMARK1;
    else if (IsNumber(tag))
	return tag + 1 - LISTNUMBER1;
    else
	return 0;
}


/*****************************************************************************
 int ListType()
 *****************************************************************************/

int ListType(tag)
      int tag;
{
    return IsNumber(tag) ? ORDERED : IsBullet(tag) ? UNORDERED : DESCRIPTION;
}


/*****************************************************************************
 void TCDescend()
 *****************************************************************************/

void TCDescend(tclevels)
      int tclevels;
{
    if (!NOTOC)
	{
	    fprintf(indexfile, "\n");
	    for ( ; tclevels; tcdepth++, tclevels--)
		fprintf(indexfile, "<UL>\n");
	}
}


/*****************************************************************************
 void TCAscend()
 *****************************************************************************/

void TCAscend(tclevels)
      int tclevels;
{
    if (!NOTOC)
	{
	    if (tclevels == 0)
		fprintf(indexfile, "\n<BR>");
	    else
		{
		    for ( ; tclevels; tcdepth--, tclevels--)
			fprintf(indexfile, "</UL>\n");
		}
	    fprintf(indexfile, "\n");
	}
}


/*****************************************************************************
 void Descend()
 *****************************************************************************/

void Descend(list_type, levels)
      int list_type;
      int levels;
{
    fprintf(f, "\n");
    for ( ; levels; stack[++depth] = list_type, levels--)
	    fprintf(f, "%s\n", list_type == ORDERED ? "<OL>" :
		    list_type == UNORDERED ? "<UL>" : "<DL>");
}


/*****************************************************************************
 void Ascend()
 *****************************************************************************/

void Ascend(levels)
      int levels;
{
    for ( ; levels; depth--, levels--)
	fprintf(f, "%s\n", stack[depth] == ORDERED ? "</OL>\n" :
		stack[depth] == UNORDERED ? "</UL>\n" : "</DL>\n");
    fprintf(f, "\n");
}



/******************************************************************************
 void PrintString() : Prints to standard output the contents of a MIF String
                      statement.
 Special cases	    : Footnotes, TOCs
 *****************************************************************************/


void PrintString()
{
    char *string_contents;

    if (!in_string)
	PrintHeaderTypeStart();
    in_string = TRUE;

    string_contents = GetContents();
    if (!string_contents)
        return;

    if (in_footnote)
	{
	    if (footnote_def_count >= MAXNUMFOOTNOTES)
		{
		    fprintf(stderr, "trans: ERROR: Too many footnotes.\n");
		    return;
		}
	    if (marker_text)
		{
		    strcat(footnote[footnote_def_count], marker_text);
		    strcat(footnote[footnote_def_count], string_contents);
		    strcat(footnote[footnote_def_count], "</A>");
		    free(marker_text);
		    marker_text = NULL;
		}	
	    else
		strcat(footnote[footnote_def_count], string_contents);
	    free(string_contents);
	}		    
    else if (header == TOC) 
	{
	    /* Remove table of contents */
	    free(string_contents);
	}
    else
	{
	    if (marker_text)
		{
		    strcat(paragraph, marker_text);
		    strcat(paragraph, string_contents);
		    strcat(paragraph, "</A>");
		    free(marker_text);
		    marker_text = NULL;
		}
	    else
		strcat(paragraph, string_contents);
	    free(string_contents);
	}
}


/******************************************************************************
 void ParagraphTag() : 
             Output : Set the current paragraph tag.
 *****************************************************************************/

void ParagraphTag()
{
    header = HeaderType();

    if (in_pgf_catalog || in_font_catalog)
        return;

    if (header == PROGRAM)
        {
	    /* djf: had to move this here from PrintHeaderTypeEnd() so that
	            in_program is set properly when deciding whether to insert
		    <CODE> tag */
	    if (!in_program)
	        {
		    in_program = TRUE;
		    fprintf(f, "<PRE>\n");
		}
	}
}


/******************************************************************************
 void FontTag() : 
             Output : Set the current font tag.
 *****************************************************************************/

void FontTag()
{
    char *ftag;
    struct FMFont *font;

    if (in_pgf_catalog)
        {
	    /* do nothing */
	}
    else if (in_font_catalog)
        {
	    current_font->tag = GetValue();
	}
    else
        {
	    ftag = GetValue();
	    if (!ftag)
	        {
		    ChangeFontWeight("");
		    ChangeFontAngle("");
		    if (in_code)
		        {
		            strcat(paragraph, "</CODE>");
			    in_code = FALSE;
			}
		}
	    else
		{
		    for (font = font_list; font != NULL; font = font->next_font)
			{
			    if (!strcmp(ftag, font->tag))
				{
				    ChangeFontWeight(font->weight);
				    ChangeFontAngle(font->angle);
				    ChangeFontUnderline(font->underline);
				    ChangeFontSubscript(font->subscript);
				    ChangeFontSuperscript(font->superscript);

				    break;
				}
			}
		    free(ftag);
		}
  	}
}


/******************************************************************************
 void PrintHeaderTypeStart() : 
 *****************************************************************************/

void PrintHeaderTypeStart()
{
    if (in_reference && (header != REFERENCE && header != COMREFERENCE))
	{
	    fprintf(f, "</DL>\n");
	    in_reference = FALSE;
	}
    else if (in_glossary && (header != in_glossary))
	{
	    /* Close-up the nested glossary */

	    switch (in_glossary)
		{
		  case GLOSSARY1:
		  case COMGLOSSARY1:
		    fprintf(f, "</DL>\n");
		    break;
		  case GLOSSARY2:
		  case COMGLOSSARY2:
		    fprintf(f, "</DL></UL>\n");
		    break;
		  case GLOSSARY3:
		  case COMGLOSSARY3:
		    fprintf(f, "</DL></UL></UL>\n");
		    break;
		    
		  default:
		    if (!QUIET)
		      fprintf(stderr, "trans: Don't know glossary state %d\n",
			      in_glossary);
		}
	    in_glossary = FALSE;
	}
    else if (in_acronym && header != ACRONYM)
	{
	    fprintf(f, "</DL>\n");
	    in_acronym = FALSE;
	}
    else if (in_program && header != PROGRAM)
	{
	    fprintf(f, "</PRE>\n");
	    in_program = FALSE;
	}
}	


/******************************************************************************
 void PrintHeaderTypeEnd() : 
 *****************************************************************************/

void PrintHeaderTypeEnd()
{
    char *marks, *marke;
    
    marks = calloc(100, sizeof(char));
    marke = calloc(10, sizeof(char));
    
    if (marker_number)
	{
	    sprintf(marks, "<A NAME=\"REF%d\">", marker_number);
	    strcpy(marke, "</A>");
	}
    
    if (in_tables)
	{
	    char *formatted_paragraph = calloc(strlen(paragraph) + 200,
					       sizeof(char));
	    sprintf(formatted_paragraph, "%s%s%s%s", marks,
		    para_num_string, paragraph, marke);
	    UseString(formatted_paragraph);
	    free(formatted_paragraph);
	}
    else
	{
	    if (depth > 0 && !IsList(header) && !IsContinue(header))
		{
		    Ascend(depth);
		}
	    
	    if (header == BODY || header == CELLHEADING || header == CELLBODY)
		{
		    fprintf(f, "<P>%s%s%s%s\n", marks, para_num_string,
			    paragraph, marke);
		}
	    else if (IsList(header))
		{
		    int relative_depth = Level(header) - depth;
		    int list_type = ListType(header);
		    
		    if (relative_depth > 0)
			{
			    Descend(list_type, relative_depth);
			}
		    else
			{
			    Ascend(-(relative_depth));
			    
			    if (list_type != stack[depth])
				{
				    Ascend(1);
				    Descend(list_type, 1);
				}
			}
		    fprintf(f, "%s%s%s%s%s\n<BR>", marks, IsBullet(header) ||
			    IsNumber(header) ? "<P><LI>" : "<DD>",
			    IsMark(header) ? para_num_string : "",
			    paragraph, marke);
		}
	    else if (IsContinue(header))
		{
		    fprintf(f, "%s%s%s\n<BR>", marks, paragraph, marke);
		}
	    else if (header == AUTHOR)
		{
		    fprintf(titlefile, "<P><ADDRESS>%s%s%s%s</ADDRESS>\n", 
			    marks, para_num_string, paragraph, marke);
		}
	    else if (header == ADDRESS)
		{
		    fprintf(f, "<P><ADDRESS>%s%s%s%s</ADDRESS>\n", 
			    marks, para_num_string, paragraph, marke);
		}
	    else if (header == TITLE)
		{
		    char *overtitle = getenv("TITLE");
		    if (!(overtitle && overtitle[0])) 
			{
			    char *tmptitle = GetTextOnly(paragraph);
			    fprintf(titlefile, "<TITLE>%s%s</TITLE>\n", 
				    para_num_string, tmptitle);
			    fprintf(titlefile, "<H1>%s%s%s%s</H1>\n", 
				    marks, para_num_string, paragraph, marke);
			    free(tmptitle);
			}
		    in_heading = FALSE;
		}
	    else if (header == LOCALTITLE)
		{
		    char *tmptitle = GetTextOnly(paragraph);
		    fprintf(f, "<TITLE>%s%s</TITLE>\n", 
			    para_num_string, tmptitle);
		    free(tmptitle);

		    if (marker_number)
			{
			    fprintf(f, "<H1><A NAME=\"REF%d\">%s%s</A></H1>\n",
				    marker_number, para_num_string,
				    paragraph);

			    if (!NOTOC)
			        SaveTitle(TRUE);
			}
		    else
			{
			    fprintf(f, "<H1><A NAME=\"HDT%d\">%s%s</A></H1>\n",
				    header_id, para_num_string,
				    paragraph);

			    if (!NOTOC)
			        SaveTitle(FALSE);

			    header_id++;
			}
		    in_heading = FALSE;
		}
	    else if (header >= HEADING0 && header <= HEADING6) 	
		{
		    int relative_depth = header - HEADING1 + 1 - tcdepth;
		    int hlevel = header - HEADING1 + 2;

		    if (!NOTOC && !title_printed)
		        PrintTitle();

		    if (hlevel > 6) hlevel = 6;

		    if (relative_depth > 0)
			{
			    TCDescend(relative_depth);
			}
		    else
			{
			    TCAscend(-(relative_depth));
			}
		    
		    if (marker_number)
			{
			    fprintf(f, "<H%d><A NAME=\"REF%d\">%s%s</A></H%d>\n", 
				    hlevel, marker_number, para_num_string, 
				    paragraph, hlevel);

			    if (!NOTOC)
				{
				    fprintf(indexfile, "<A HREF=\"%s.html#REF%d\">%s%s</A>\n",
					    TRANSBasename, marker_number, 
					    para_num_string, paragraph);
				}
			}
		    else
			{
			    fprintf(f, "<H%d><A NAME=\"HDR%d\">%s%s</A></H%d>\n", 
				    hlevel, header_id,
				    para_num_string, paragraph, hlevel);

			    if (!NOTOC)
				{
				    fprintf(indexfile, "<A HREF=\"%s.html#HDR%d\">%s%s</A>\n",
					    TRANSBasename, header_id,
					    para_num_string, paragraph);
				}
			    header_id++;
			}
		    in_heading = FALSE;
		}
	    else if (header >= UHEADING0 && header <= UHEADING6) 
		{
		    int hlevel = header - UHEADING1 + 2;
		    if (hlevel > 6)
		        hlevel = 6;
		    fprintf(f, "<H%d>%s%s%s%s</H%d>\n", 
			    hlevel, marks, para_num_string, paragraph,
			    marke, hlevel);
		    in_heading = FALSE;
		}
	    else if (header >= INDENTHEADING0 && header <= INDENTHEADING6) 
		{
		    int hlevel = header - INDENTHEADING1 + 2;
		    if (hlevel > 6)
		        hlevel = 6;
		    fprintf(f, "<UL><H%d>%s%s%s%s</H%d></UL>\n", 
			    hlevel, marks, para_num_string, paragraph,
			    marke, hlevel);
		    in_heading = FALSE;
		}
	    else if (header == TABLE || header == FIGURE) 
		{
		    fprintf(f, "<P>%s%s%s%s\n", marks, para_num_string,
			    paragraph, marke);
		}
	    else if (header == ABSTRACT)
		{
		    fprintf(f, "<P><EM>%s%s%s%s</EM>\n", marks,
			    para_num_string, paragraph, marke);
		}    
	    else if (header == REFERENCE || header == COMREFERENCE) 
		{
		    if (!in_reference)
			{
			    if (header == REFERENCE)
				fprintf(f, "<DL>\n");
			    else
				fprintf(f, "<DL COMPACT>\n");
			    in_reference = TRUE;
			}
		    fprintf(f, "<DT>%s%s%s<DD>%s\n", marks, para_num_string,
			    marke, paragraph);     
		}
	    else if (header >= GLOSSARY1 &&  header <= COMGLOSSARY3) 
		{
		    if (!in_glossary)
			{
			    in_glossary = header;
			    if (header >= GLOSSARY1 && header <= GLOSSARY3)
				{
				    int i;
				    for (i = 1; i <= header - GLOSSARY1; i++)
					fprintf(f, "<UL>");
				    fprintf(f, "<DL>\n");
				}
			    else
				{
				    int i;
				    for (i = 1; i <= header - COMGLOSSARY1;i++)
					fprintf(f, "<UL>");
				    fprintf(f, "<DL COMPACT>\n");
				}
			}
		    fprintf(f, "<DT>%s%s%s%s\n", marks, para_num_string,
			    paragraph, marke);
		}
	    else if (header == PROGRAM)
		{
		    fprintf(f, "%s%s%s%s\n", marks, para_num_string,
			    paragraph, marke);
		}
	    else if (header == ACRONYM) 
		{
		    if (!in_acronym)
			{
			    in_acronym = TRUE;
			    fprintf(f, "<DL COMPACT>\n");
			}
		    fprintf(f, "<DT>%s%s%s%s\n", marks, para_num_string,
			    paragraph,
			    marke);
		}
           else if (header == TOC)	/* ejr: Don't break flow */
        	{
        	}
	   else if (header >= HTML) 	/* mnh: HTML headers */
		{
		    char tmp[100];
		    char *cp;
		    strcpy(tmp, html_tags[header-HTML]);
		    cp = tmp;
		    while (*cp && (*cp != ';'))
		        cp++;
		    if (*cp) 
			{ 
			    *cp++ = '\0';
			    fprintf(f, "%s%s%s%s%s%s\n", tmp, marks, 
				    para_num_string, paragraph, marke, cp);
			} 
		    else 
			{
			    fprintf(f, "%s%s%s%s%s\n", tmp, marks, 
				    para_num_string, paragraph, marke);
			} 
		}
	    else 
		{
		    fprintf(f, "<UNKNOWN>%s%s%s%s</UNKNOWN>\n", 
			    marks, para_num_string, paragraph, marke);
		}
	}

    if (separator_pending)
        {
            fprintf(f,"\n<HR>\n");
	    separator_pending = FALSE;
	}

    paragraph[0] = 0;
    para_num_font[0] = 0;
    para_num_string[0] = 0;
    marker_number = 0;
    free(marks);
    free(marke);
}


/******************************************************************************
 int SaveTitle() : Save a title to put in table of contents later.  The idea
		   is to use the first instance of a LOCALTITLE paragraph which
		   occurs on a body page.  This is because there may be several
		   instances of LOCALTITLE on various master pages, e.g. in a
		   title page document which includes master pages for TOC, LOT
		   etc.  If there is no body page title then the first
		   instance of any other title is used.  (djf)
 *****************************************************************************/

int SaveTitle(is_marker)
      int is_marker;
{
    static int got_body_page_title = FALSE;

    if (got_body_page_title)
        {
	    if (!QUIET)
	        fprintf(stderr, "trans: WARNING: only one LOCALTITLE paragraph is allowed; ignoring title \"%s\"\n", paragraph);
	    return;
	}

    if (!html_title_string || in_body_page)
	{
	    if (in_body_page)
	        got_body_page_title = TRUE;

	    if (html_title_string)
		free(html_title_string);

	    html_title_string = malloc(strlen(TRANSBasename) +
				       strlen(para_num_string) +
				       strlen(paragraph) + 100);

	    if (is_marker)
		sprintf(html_title_string,
			"<P><A HREF=\"%s.html#REF%d\">%s%s</A>\n",
			TRANSBasename, marker_number,
			para_num_string, paragraph);
	    else
		sprintf(html_title_string,
			"<P><A HREF=\"%s.html#HDT%d\">%s%s</A>\n",
			TRANSBasename, header_id,
			para_num_string, paragraph);
	}
}


/******************************************************************************
 int PrintTitle() : Write the HTML title string to the index file.  This is
		    done immediately before the first HEADING is written, or
		    if there are no HEADINGs, at the end of processing
 *****************************************************************************/

int PrintTitle()
{
    if (html_title_string)
        {
	    fprintf(indexfile, "%s\n", html_title_string);
	    title_printed = TRUE;
	}
}


/******************************************************************************
 int HeaderType() : 
 *****************************************************************************/

int HeaderType()
{
    int header_type = UNKNOWN;
    char *pgftag;
    int i;

    pgftag = GetValue();
  
    if (!pgftag) 
        return header_type;

    for (i = 0; i < tags_count && strcmp(fm_tags[i], pgftag); i++)
	;

    if (i < tags_count)
        {
	    header_type = int_tags[i];
	}
    else if (strstr(pgftag, "TOC") || strstr(pgftag, "IX") ||
	     strstr(pgftag, "LOT") || strstr(pgftag, "LOF") ||
	     strstr(pgftag, "AIX") || strstr(pgftag, "AML") ||
	     strstr(pgftag, "APL") || strstr(pgftag, "IOM") ||
	     strstr(pgftag, "LOM") || strstr(pgftag, "LOP") ||
	     strstr(pgftag, "SIX"))
        {
	    header_type = TOC;
	}
    else
        {
	    int j;

	    if (!in_pgf_catalog)
		{
		    for (j = 0; j < unknown_ext_tags_count; j++)
			if (strcmp(pgftag, unknown_ext_tags[j]) == 0)
			    break;

		    if (j == unknown_ext_tags_count)
			{
			    unknown_ext_tags[unknown_ext_tags_count++] =
			        strdup(pgftag);
			    fprintf(stderr,
				    "trans: Unknown paragraph tag: \"%s\"\n",
				    pgftag);
			}
		}
        }

    if (in_pgf_catalog)
	{
	    current_pgf->tag = strdup(pgftag);
	}
    else
	{
	    if (header_type == TITLE || header_type == LOCALTITLE ||
		(header_type >= HEADING0 && header_type <= HEADING6) ||
		(header_type >= UHEADING0 && header_type <= UHEADING6) ||
		(header_type >= INDENTHEADING0 &&
		 header_type <= INDENTHEADING6))
	        in_heading = TRUE;

	    for (current_pgf = pgf_list; current_pgf != NULL;
		 current_pgf = current_pgf->next_pgf)
		{
		    if (!strcmp(current_pgf->tag, pgftag))
			{
			    if (current_cell)
			        current_cell->alignment = current_pgf->alignment;

			    if (current_pgf->topseparator)
				fprintf(f,"<HR>\n");

			    if (current_pgf->botseparator)
			        separator_pending = TRUE;

			    ChangeFontWeight(current_pgf->font->weight);
			    ChangeFontAngle(current_pgf->font->angle);
			    ChangeFontUnderline(current_pgf->font->underline);
			    ChangeFontSubscript(current_pgf->font->subscript);
			    ChangeFontSuperscript(current_pgf->font->superscript);
			    break;
			}
		}
	}

    free(pgftag);
    return header_type;
}


/******************************************************************************
 void StartPrint() : 
 *****************************************************************************/

void StartPrint()
{
    int j, k, c;
    char tag[100];
    char *overtitle = getenv("TITLE");
    int comment = FALSE;
    int found = FALSE;

    debug("StartPrint\n");

    if (!QUIET)
        fprintf(stderr, "\n");

    TRANSFilename = getenv("TRANSFilename");
    TITLEFilename = getenv("TITLEFilename");
    INDEXFilename = getenv("INDEXFilename");
    HEADFilename  = getenv("HEADFilename");
    TAGSFilename  = getenv("TAGSFilename");
    XDPI	  = getenv("XDPI");
    YDPI	  = getenv("YDPI");
    NOCROP	  = getenv("NOCROP");
    NOGRAPHICS    = getenv("NOGRAPHICS");
    NOTOC	  = getenv("NOTOC");
    STRIP	  = getenv("STRIP");
    QUIET	  = getenv("QUIET");
    DEBUG	  = getenv("DEBUG");

    if (TRANSBasename = strrchr(TRANSFilename, '/'))
        TRANSBasename += 1;
    else
        TRANSBasename = TRANSFilename;

    strcpy(pagefilename, TRANSFilename);
    strcat(pagefilename, ".tmp.000P");

    if (!(pagefile = fopen(pagefilename, "w")))
	{
	    fprintf(stderr,
		    "trans: ERROR: Could not open temporary file \"%s\"\n",
		    pagefilename);
	    exit(-1);
	}

    strcpy(flowfilename, TRANSFilename);
    strcat(flowfilename, ".tmp.000F");

    if (!(flowfile = fopen(flowfilename, "w")))
	{
	    fprintf(stderr,
		    "trans: ERROR: Could not open temporary file \"%s\"\n",
		    flowfilename);
	    exit(-1);
	}

    if (!NOTOC)
	{
	    if (!(indexfile = fopen(INDEXFilename, "a")))
		{
		    fprintf(stderr,
			    "trans: ERROR: Could not open temporary file \"%s\"\n",
			    INDEXFilename);
		    exit(-1);
		}
	}

    if (!(titlefile = fopen(TITLEFilename, "a")))
	{
	    fprintf(stderr,
		    "trans: ERROR: Could not open temporary file \"%s\"\n",
		    TITLEFilename);
	    exit(-1);
	}

    if (!(tagsfile = fopen(TAGSFilename, "r")))
	{
	    fprintf(stderr,
		    "trans: ERROR: Could not open tags file \"%s\"\n",
		    TAGSFilename);
	    exit(-1);
	}

    /* Print header block */

    f = pagefile;

    fprintf(f, "<!-- This file was created with the fm2html filter.\n");
    fprintf(f, "The filter is copyright Norwegian Telecom Research and\n");
    fprintf(f, "was programmed by Jon Stephenson von Tetzchner. -->\n\n");

    if (overtitle && overtitle[0]) 
        fprintf(f, "<TITLE>%s</TITLE>\n", overtitle);

    /* Fill tags array */
    
    j = 0;
    k = 0;

    while ((c=getc(tagsfile)) != EOF) 
	{
	    if (c == '\t' && !found) 
		{
		    if (k > 0) 
			{
			    tag[k] = '\0';
			    StripSpace(tag);
			    k = 0;
			    if (strlen(tag))
				{	
				    strcpy(fm_tags[tags_count],tag);
				    found = TRUE;
				}
			}
		} 
	    else if (c == '\n' || c == '#' || c=='\t') 	
		{
		    if (k > 0)
			{
			    tag[k] = '\0';
			    StripSpace(tag);
			    k = 0;
			    if (strlen(tag))
				{
				    int_tags[tags_count++] = GetIntTag(tag);
				    found = FALSE;
				}
			}
		    if (c == '#')
			{
			    comment = TRUE;
			    found = FALSE;
			}
		    else if (c == '\n')
			{
			    comment = FALSE;
			    found = FALSE;
			}
		} 
	    else if (!comment)
		{
		    tag[k++] = c;
		}
	}

    para_num_font = malloc(50);
    para_num_font[0] = 0;
    para_num_string = malloc(100);
    para_num_string[0] = 0;
    text_flow_tag = malloc(50);
    text_flow_tag[0] = 0;
    reference_filename = malloc(100);
    reference_filename[0] = 0;
    paragraph[0] = 0;
}


/******************************************************************************
 void EndPrint() : 
 *****************************************************************************/

void EndPrint()
{
    struct stat stbuf;
    int length;
    int i;

    debug("EndPrint\n");

    if (font_bold)
	fprintf(flowfile, "</B>");
    if (font_italic)
	fprintf(flowfile, "</I>");
    if (font_underline)
	fprintf(flowfile, "</U>");
    if (font_subscript)
	fprintf(flowfile, "</EM>");
    if (font_superscript)
	fprintf(flowfile, "</EM>");
    if (in_code)
	fprintf(flowfile, "</CODE>");

    if (footnote_def_count > 0)
        PrintFootnotes();
    if (tcdepth)
        TCAscend(tcdepth);
    if (!NOTOC && !title_printed)
        PrintTitle();

    fprintf(flowfile, "\n<A NAME=\"ENDFILE\"><PRE> </PRE></A>\n");
    fprintf(flowfile, "\n<<<\n");
    fclose(flowfile);

    fclose(pagefile);
    
    if (!NOTOC)
        fclose(indexfile);

    if (NOGRAPHICS)
	return;

    /* djf: Translate MIF figures to PostScript and GIF, and MIF tables to
            PostScript.  Have to be careful not to let arguments to mif2gif
	    or mif2ps get so long that they are truncated.  I think 1024
	    characters should be universally safe but one never know, do
	    one.  Alternative would be to translate one file at a time,
	    but that would be much slower. */

    for (i = 0; i < frame_count; i++)
	{
	    sprintf(sys_call, "cat %s.tmp >> %s",
		    frame_filenames[i], frame_filenames[i]);
	    system(sys_call);
	    if (!DEBUG)
	        {
		    sprintf(sys_call, "rm -f %s.tmp", frame_filenames[i]);
		    system(sys_call);
		}
	}

    for (i = 0; i < frame_count; )
	{
	    sprintf(sys_call, "-SCRDIR-/mif2gif %s %s %s %s %s %s",
		    NOCROP ? "" : "-c", XDPI ? XDPI : "", YDPI ? YDPI : "",
		    QUIET ? "-q" : "", DEBUG ? "" : "-r", frame_filenames[i]);
	    length = strlen(sys_call);

	    for (i++; i < frame_count; i++)
		{
		    length += (1 + strlen(frame_filenames[i]));
		    if (length > 1024)
		        break;
		    strcat(sys_call, " ");
		    strcat(sys_call, frame_filenames[i]);
		}
	    system(sys_call);
	}	    

    for (i = 0; i < table_count; )
	{
	    sprintf(sys_call, "-SCRDIR-/mif2ps %s %s %s",
		    QUIET ? "-q" : "", DEBUG ? "" : "-q", table_filenames[i]);
	    length = strlen(sys_call);
	    for (i++; i < table_count; i++)
		{
		    length += (1 + strlen(table_filenames[i]));
		    if (length > 1024)
		        break;
		    strcat(sys_call, " ");
		    strcat(sys_call, table_filenames[i]);
		}
	    system(sys_call);
	}

    /* Remove empty GIFs (have length 35 or 0) */

    for (i = 0; i < frame_count; i++)
        {
	    strcpy(frame_filename, frame_filenames[i]);
	    frame_filename[strlen(frame_filename) - 3] = 'g';

	    if (stat(frame_filename, &stbuf) == 0)
	        if (stbuf.st_size == 35 || stbuf.st_size == 0) 
		    {
		        if (!QUIET)
			    fprintf(stderr,
				    "trans: Deleting empty GIF file: %s\n",
				    frame_filename);
			unlink(frame_filename);
		    }
	}

    for (i = 0; i < table_count; i++)
        {
	    strcpy(table_filename, table_filenames[i]);
	    table_filename[strlen(table_filename) - 3] = 'g';

	    if (stat(table_filename, &stbuf) == 0)
	        if (stbuf.st_size == 35 || stbuf.st_size == 0) 
		    {
		        if (!QUIET)
			    fprintf(stderr,
				    "trans: Deleting empty GIF file: %s\n",
				    table_filename);
			unlink(table_filename);
		    }
	}
}


/******************************************************************************
 void PageStart() : 
 *****************************************************************************/

void PageStart()
{
    debug("PageStart\n");
}


/******************************************************************************
 void PageType() : 
 *****************************************************************************/

void PageType()
{
    debug("PageType\n");

    if (strstr(yytext, "BodyPage"))
        in_body_page = TRUE;
}


/******************************************************************************
 void PageEnd() : 
 *****************************************************************************/

void PageEnd()
{
    debug("PageEnd\n");

    in_body_page = FALSE;
}


/******************************************************************************
 void ImportObjectStart() : 
 *****************************************************************************/

void ImportObjectStart()
{
    debug("ImportObjectStart\n");

    if (in_body_page)
        {
            struct FMImportObject *new_impobj = (struct FMImportObject *)
	        calloc(1, sizeof(struct FMImportObject));

	    new_impobj->textrectid = current_textrect_id;
	    new_impobj->importobfile = NULL;
	    new_impobj->importobfiledi = NULL;
	    new_impobj->brect = NULL;
	    new_impobj->angle = NULL;
	    new_impobj->fliplr = NULL;
	    new_impobj->next_importobject = NULL;

	    if (!impobj_list)
		impobj_list = new_impobj;
	    else
		current_impobj->next_importobject = new_impobj;

	    current_impobj = new_impobj;
	}
}


/******************************************************************************
 void ImportObFile() : 
 *****************************************************************************/

void ImportObFile()
{
    debug("ImportObFile\n");

    if (in_body_page)
        current_impobj->importobfile = strdup(yytext);
}


/******************************************************************************
 void ImportObFileDI() : 
 *****************************************************************************/

void ImportObFileDI()
{
    debug("ImportObFileDI\n");

    if (in_body_page)
        current_impobj->importobfiledi = strdup(yytext);
}


/******************************************************************************
 void BRect() : 
 *****************************************************************************/

void BRect()
{
    debug("BRect\n");

    if (in_body_page)
        current_impobj->brect = strdup(yytext);
}


/******************************************************************************
 void Angle() : 
 *****************************************************************************/

void Angle()
{
    debug("Angle\n");

    if (in_body_page)
        current_impobj->angle = strdup(yytext);
}


/******************************************************************************
 void FlipLR() : 
 *****************************************************************************/

void FlipLR()
{
    debug("FlipLR\n");

    if (in_body_page)
        current_impobj->fliplr = strdup(yytext);
}


/******************************************************************************
 void ImportObjectEnd() : 
 *****************************************************************************/

void ImportObjectEnd()
{
    debug("ImportObjectEnd\n");
}


/******************************************************************************
 void InsertFootnote() : 
 *****************************************************************************/

void InsertFootnote()
{
    char sss[350]; 

    if (!in_string)
	PrintHeaderTypeStart();
    in_string = TRUE;
	    
    footnote_use_count++;
    if (in_tables)
	{
	    sprintf(sss, "<A HREF=\"%s.html#FN%d%c\">(%c)</A>", 
		    TRANSBasename, current_table->id, 
		    footnote_use_count + 'a' - 1, footnote_use_count + 'a' - 1);
	}
    else
	{
	    sprintf(sss, "<A HREF=\"%s.html#FN%d\">(%d)</A>", TRANSBasename,
		    footnote_use_count, footnote_use_count);
	}
    strcat(paragraph, sss); 
}


/******************************************************************************
 void PrintFootnotes() : 
 *****************************************************************************/

void PrintFootnotes()
{
    int i;
    char *language = getenv("DOCUMENTLANGUAGE");

    if (language && language == "NO")	
	fprintf(pagefile, "\n<HR><H3>Fotnoter</H3>\n");
    else if (language && language == "D")	
	fprintf(pagefile, "\n<HR><H3>Fu&szlig;noten</H3>\n"); 
    else
	fprintf(pagefile, "\n<HR><H3>Footnotes</H3>\n");
    fprintf(pagefile, "<DL COMPACT>\n");

    for (i = 1; i <= footnote_def_count; i++)
	{	
	    fprintf(pagefile, "<DT><A NAME=FN%d>(%d)</A><DD>%s\n", i, i, footnote[i]);
	    free(footnote[i]);
	}

    fprintf(pagefile, "</DL>\n");
    footnote_def_count = 0;
}


/******************************************************************************
 void NewFootnoteStart() : 
 *****************************************************************************/

void NewFootnoteStart()
{
    footnote_def_count++;
    footnote[footnote_def_count] = calloc(1000, 1);    
    footnote[footnote_def_count][0] = 0;
    in_footnote = TRUE;
}


/******************************************************************************
 void NewFootnoteEnd() : 
 *****************************************************************************/

void NewFootnoteEnd()
{
    in_footnote = FALSE;
}


/******************************************************************************
 void InsertReferenceStart() : 
 *****************************************************************************/

void InsertReferenceStart()
{
}


/******************************************************************************
 void SetReferenceNumber() : 
 *****************************************************************************/

void SetReferenceNumber()
{
    reference_number = GetNumber();
}

/******************************************************************************
 void SetReferenceFile() : 
 *****************************************************************************/

void SetReferenceFile()
{
    char ref_string[350];

    GetReferenceFile();
    sprintf(ref_string, "<A HREF=\"%s#REF%d\">", reference_filename,
	    reference_number);
    if (!in_string)
	PrintHeaderTypeStart();
    in_string = TRUE;
    strcat(paragraph, ref_string);
}

/******************************************************************************
 void InsertReferenceEnd() : 
 *****************************************************************************/

void InsertReferenceEnd()
{
    if (!in_string)
	PrintHeaderTypeStart();
    in_string = TRUE;
    strcat(paragraph, "</A>");
}


/******************************************************************************
  void PgfCatalogStart() :
        Input  : 
	Output : 
 *****************************************************************************/
void PgfCatalogStart()
{
    debug("PgfCatalogStart\n");

    in_pgf_catalog = TRUE;
}


/******************************************************************************
  void PgfCatalogEnd() :
        Input  : 
	Output : 
 *****************************************************************************/
void PgfCatalogEnd()
{
    debug("PgfCatalogEnd\n");

    in_pgf_catalog = FALSE;
}


/******************************************************************************
  void FontCatalogStart() :
        Input  : 
	Output : 
 *****************************************************************************/
void FontCatalogStart()
{
    debug("FontCatalogStart\n");

    in_font_catalog = TRUE;
}


/******************************************************************************
  void FontCatalogEnd() :
        Input  : 
	Output : 
 *****************************************************************************/
void FontCatalogEnd()
{
    debug("FontCatalogEnd\n");

    in_font_catalog = FALSE;
}


/******************************************************************************
  void AFramesStart() : 
        Input  : 
	Output : 
 *****************************************************************************/
void AFramesStart()
{
    debug("AFramesStart\n");
}


/******************************************************************************
  void AFramesEnd() :
        Input  : 
	Output : 
 *****************************************************************************/

void AFramesEnd()
{
    debug("AFramesEnd\n");

    aframe_count = frame_count;
}


/******************************************************************************
  void FrameStart() : 
        Input  : 
	Output : 
 *****************************************************************************/

void FrameStart()
{
    debug("FrameStart\n");

    in_frame = TRUE;
    textrect_count = 0;
    current_frame_id = GetIDNextLine();

    if (NOGRAPHICS)
	return;

    sprintf(frame_filename, "%s.fig_%d.mif", TRANSFilename, current_frame_id);
    sprintf(sys_call, "cp %s %s", HEADFilename, frame_filename);
    if (!QUIET)
        fprintf(stderr, "trans: Making FrameMaker MIF file \"%s\"\n",
		frame_filename);
    system(sys_call);

    framef = fopen(frame_filename, "a");
    if (!framef)
	{
	    fprintf(stderr, "trans: ERROR: Could not open file \"%s\"\n",
		    frame_filename);
	    exit(-1);
	}
}


/******************************************************************************
  void FrameEnd() :
        Input  : 
	Output :
 *****************************************************************************/

void FrameEnd()
{
    int i;

    debug("FrameEnd\n");

    in_frame = FALSE;

    if (NOGRAPHICS)
	return;

    fclose(framef);

    sprintf(frame_tmpname, "%s.tmp", frame_filename);
    frametmpf = fopen(frame_tmpname, "w");
    if (!frametmpf)
	{
	    fprintf(stderr, "trans: ERROR: Could not open file \"%s\"\n",
		    frame_tmpname);
	    exit(-1);
	}
    fprintf(frametmpf, "<TextFlow \n");
    fprintf(frametmpf, " <Para \n");
    fprintf(frametmpf, "  <PgfTag `fm2htmlparagraphformat'>\n");
    fprintf(frametmpf, "  <ParaLine \n");
    fprintf(frametmpf, "   <AFrame %d>\n", current_frame_id);
    fprintf(frametmpf, "  >\n");
    fprintf(frametmpf, " > # end of Para\n");
    fprintf(frametmpf, "> # end of TextFlow\n");
    fclose(frametmpf);

    frame_filenames[frame_count++] = (char *) strdup(frame_filename);

    if (textrect_count > 0)
	{
	    sprintf(sys_call, "-SCRDIR-/addtextrect %s %s", TRANSFilename,
		    frame_tmpname);

	    for (i = 0; i < textrect_count; i++) 
		{
		    char temp_num[10];
		    strcat(sys_call, " ");
		    sprintf(temp_num, " %d%c", textrect_ids[i], 0);
		    strcat(sys_call, temp_num);
		}
	    system(sys_call);
	}
}


/******************************************************************************
  void AFrameUse() :
        Input  : 
	Output : 
 *****************************************************************************/

void AFrameUse()
{
    int frame_id, frame_type;

    debug("AFrameUse\n");

    frame_id = GetID();
    frame_type = GetFrameType(frame_id);
    if (frame_type == INLINEFRAME || in_tables)
	{
	    InsertInlineFigure(f, frame_id);
	}
    else
	{
	    tabfig_ids[tabfig_count] = GetID();
	    tabfig[tabfig_count++] = FIGURE;
	}

    if (NOGRAPHICS)
        return;

    for (current_impobj = impobj_list; current_impobj != NULL;
	 current_impobj = current_impobj->next_importobject)
        {
	    if (current_impobj->textrectid == current_textrect_id)
	        {
		    sprintf(frame_filename, "%s.fig_%d.mif", TRANSFilename, 
			    frame_id);
		    framef = fopen(frame_filename, "a");

		    if (!framef)
		        {
			    fprintf(stderr, "trans: ERROR: Could not reopen file \"%s\"\n", frame_filename);
			    continue;
			}

		    fprintf(framef, "<ImportObject \n");
		    if (current_impobj->importobfile)
		        fprintf(framef, " %s\n", current_impobj->importobfile);
		    if (current_impobj->importobfiledi)
		        fprintf(framef, " %s\n", current_impobj->importobfiledi);
		    if (current_impobj->brect)
		        fprintf(framef, " %s\n", current_impobj->brect);
		    if (current_impobj->angle)
		        fprintf(framef, " %s\n", current_impobj->angle);
		    if (current_impobj->fliplr)
		        fprintf(framef, " %s\n", current_impobj->fliplr);
		    fprintf(framef, "> # end of ImportObject\n");
		    fclose(framef);
		}
	}
}


/******************************************************************************
  void TablesStart() : All tables are skipped
        Input  : 
	Output : Indicate parsing tables.
 *****************************************************************************/

void TablesStart()
{
    in_tables = TRUE;

    debug("TablesStart\n");

    if (!QUIET)
        fprintf(stderr, "\n");
}


/******************************************************************************
  void TablesEnd() :
        Input  : 
	Output : No more tables.
 *****************************************************************************/

void TablesEnd()
{
    debug("TablesEnd\n");

    in_tables = FALSE;
}


/******************************************************************************
  void TableUse() :
        Input  : 
	Output : Prints information concerning missing table.
 *****************************************************************************/

void TableUse()
{
    tabfig_ids[tabfig_count] = GetID();
    tabfig[tabfig_count++] = TABLE;
}


/******************************************************************************
  void SetMarker() :
        Input  : Uses yytext
	Output : marker_number = ID of marker
	OR for hypertext: marker_text.
 *****************************************************************************/

void SetMarker()
{
    if (marker_type == 8)       /* HyperText link */
        {
            char *content;
            content = GetContents();
            if (content)
                {
                    char *hypcommand;	/* Used to get tokens from string */
		    char *hypfilename;
		    char *hyplink;
                    char *c;            /* Pointer to ':' in linkspec */

                    /*
                     * Parse the tokens out of the string looking for
                     * openlink or gotolink
                     */

                    hypcommand = strtok(content, " \t");

                    if ((!strcasecmp(hypcommand, "openlink")) ||
                        (!strcasecmp(hypcommand, "gotolink"))) 
			{
			    hypfilename = (char *)strtok(NULL, ":");
			    if (!hypfilename)
				{
				    if (!QUIET)
				        fprintf(stderr, "trans: Hypertext \"%s\" with no parameters or space. Removed.\n", hypcommand);
				}
			    else
				{
				    hyplink = strtok(NULL, ":");
				    if (hyplink)					/* Openlink or gotolink with file and linkspec */
					{
					    marker_text = calloc(1, 1000);
					    if (c = (strstr(hypfilename, ".doc"))) 		/* Remove .doc */
						{
						    *c = 0;
						}
					    if (!strcasecmp(hyplink, "firstpage"))
						{
						    sprintf(marker_text, "<A HREF=\"%s.html\">", hypfilename);
						}
					    else if (!strcasecmp(hyplink, "lastpage"))
						{
						    sprintf(marker_text, "<A HREF=\"%s.html#ENDFILE\">", hypfilename);
						}
					    else
						{
						    sprintf(marker_text, "<A HREF=\"%s.html#%s\">", hypfilename, hyplink);
						}
					}
				    else 						/* Openlink or gotolink with only a linkspec */
					{
					    marker_text = calloc(1, 1000);
					    hyplink = hypfilename;
					    if (!strcasecmp(hyplink, "firstpage"))
						{
						    sprintf(marker_text, "<A HREF=\"%s.html\">", TRANSBasename);
						}
					    else if (!strcasecmp(hyplink, "lastpage"))
						{
						    sprintf(marker_text, "<A HREF=\"%s.html#ENDFILE\">", TRANSBasename);
						}
					    else
						{
						    sprintf(marker_text, "<A HREF=\"%s.html#%s\">", TRANSBasename, hyplink);
						}
					}
				}
			}
		    else if (!strcasecmp(hypcommand, "newlink"))
			{
			    hyplink = (char *)strtok(NULL, ":");
			    if (hyplink)
				{		
				    marker_text = calloc(1, 1000);
				    sprintf(marker_text, "<A NAME=\"%s\">", hyplink);
				}
			    else
			        if (!QUIET)
				    fprintf(stderr, "trans: FrameMaker hypertext command \"newlink\" with no parameters. Ignored.");
			}
                    else if ((!strcasecmp(hypcommand, "alert")) ||
			     (!strcasecmp(hypcommand, "matrix")) ||
			     (!strcasecmp(hypcommand, "nextpage")) ||
			     (!strcasecmp(hypcommand, "popup")) ||
			     (!strcasecmp(hypcommand, "previouslink")) ||
			     (!strcasecmp(hypcommand, "previouspage")) ||
			     (!strcasecmp(hypcommand, "quit")) ||
			     (!strcasecmp(hypcommand, "quitall")) ||
			     (!strcasecmp(hypcommand, "message")))
			{
			    if (!QUIET)
			        fprintf(stderr, "trans: The FrameMaker hypertext command \"%s\" is not converted.\n", hypcommand);
			}
		    else
			{
			    /* 
			     * Not a openlink or a gotolink, use the whole
			     * thing as a URL.
			     */
			    marker_text = calloc(1, 1000);
			    sprintf(marker_text, "<A HREF=\"%s\">", content);
			}
                    free(content);
                }
        }
    else if (marker_type == 9)
        marker_number = GetNumber();
    else
	{
	    if (!QUIET)
	        fprintf(stderr, "trans: Ignored marker of type %d.\n", marker_type); 
	}
}



/******************************************************************************
  void SetMarkerType() :
        Input  : Uses yytext
	Output : marker_type = type of marker
 *****************************************************************************/

void SetMarkerType()
{
    marker_type = atoi(yytext + 6);
}



/******************************************************************************
  char *ChangeToSymbol (str) :
 *****************************************************************************/

char *ChangeToSymbol(str)
      char *str;
{
    int i = 0;
    int j = 0;
    long val;
    char *new_str = calloc(strlen(str) * 5 + 1, sizeof(char) );
    while (str[i]) 
	{
	    val = str[i];
	    if (str[i]=='\\' && str[i+1] == 'x' && str[i+4] == ' ')
		{
		    new_str[j++] = str[i++];
		    new_str[j++] = str[i++];
		    new_str[j++] = str[i++];
		    new_str[j++] = str[i++];
		    new_str[j++] = '\\';
		    i++;
		}
	    else
		{
		    char *tempc = calloc(10,1);
		    sprintf(tempc, "\\x%x\\", val);
		    strcat(new_str, tempc);
		    free(tempc);
		    j+=5;
		    i++;
		}
	}

    return(new_str);
}


/******************************************************************************
  char *GetContents() :
        Input  : Uses yytext
	Output : Text in yytext between ` '
 *****************************************************************************/

char *GetContents()
{
    int i = 0, j = 0;
    char *quoted;

    quoted = malloc(strlen(yytext) * 5 );  /* To be able to handle < > and & */

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    i++;
	    while (yytext[i] && yytext[i] != 39)
		{
		    if (yytext[i] == '<') 
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'l';
			    quoted[j++] = 't';
			    quoted[j++] = ';';
			}
		    else if (yytext[i] == '\\' && yytext[i+1] == '>')
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'g';
			    quoted[j++] = 't';
			    quoted[j++] = ';';		
			    i++;
			}
		    else if (yytext[i] == '&')
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'a';
			    quoted[j++] = 'm';
			    quoted[j++] = 'p';
			    quoted[j++] = ';';
			}
                   else if (yytext[i] == '\\' && yytext[i+1] == 'q')
                       {
                           /* ejr: handle smart quote \q */
                           quoted[j++] = '\'';
                           i++;
                       }
                   else if (yytext[i] == '\\' && yytext[i+1] == '\\')
                       {
                           /* ejr: handle escaped slash \\ */
                           quoted[j++] = '\\';
                           i++;
                       }
		    else		    
			quoted[j++] = yytext[i];
		    i++;
		}	
	    if (yytext[i] == 39)
		{
		    quoted[j] = 0;
		    if (font_type == SYMBOL)
			return ChangeToSymbol(quoted);
		    else
			return quoted;
		}
	    else
		{
		    free(quoted);
		    return NULL;
		}
	}			
    else 
	{
	    free(quoted);
	    return NULL;
	}
}	


/******************************************************************************
  char *GetValue() :
        Input  : Uses yytext
	Output : Text in yytext between ` ', or NULL if none
 *****************************************************************************/

char *GetValue()
{
    int i = 0, j = 0;
    char *quoted;

    quoted = malloc(strlen(yytext) * 5 );  /* To be able to handle < > and & */

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    i++;
	    if (yytext[i] == 39)
		{
		    free(quoted);
		    return NULL;
		}
	    while(yytext[i] && yytext[i] != 39)
		{
		    if (yytext[i] == '<') 
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'l';
			    quoted[j++] = 't';
			    quoted[j++] = ';';
			}
		    else if (yytext[i] == '\\' && yytext[i+1] == '>')
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'g';
			    quoted[j++] = 't';
			    quoted[j++] = ';';		
			    i++;
			}
		    else if (yytext[i] == '&')
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'a';
			    quoted[j++] = 'm';
			    quoted[j++] = 'p';
			    quoted[j++] = ';';
			}
		    else		    
			quoted[j++] = yytext[i];
		    i++;
		}	
	    if (yytext[i] == 39)
		{
		    quoted[j] = 0;
		    return quoted;
		}
	    else
		{
		    free(quoted);
		    return NULL;
		}
	}			
    else 
	{
	    free(quoted);
	    return NULL;
	}
}	


/******************************************************************************
  int GetNumber() :
        Input  : Uses yytext
	Output : Number identifying marker
 *****************************************************************************/

int GetNumber()
{
    int i = 0;

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    i++;
	    return atoi(&(yytext[i]));
	}			
    else return 0;
}


/******************************************************************************
  ParagraphNumberFont() :
        Input  : Uses yytext
	Output : Paragraph numbering font string
 *****************************************************************************/

void ParagraphNumberFont()
{
    int i = 0, j = 0;

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    i++;
	    while (yytext[i] && yytext[i] != 39)
	        para_num_font[j++] = yytext[i++];
	    para_num_font[j] = 0;
	}			
    else
        para_num_font[0] = 0;

    if (*para_num_font && in_pgf_catalog)
        current_pgf->pgfnumfontstr = strdup(para_num_font);
}


/******************************************************************************
  ParagraphNumberString() :
        Input  : Uses yytext
	Output : Paragraph number string with \t removed, but with
		 character formatting tags inserted as appropriate
 *****************************************************************************/

void ParagraphNumberString()
{
    int i = 0, j = 0;
    int is_bold = FALSE;
    int is_italic = FALSE;
    int is_underline = FALSE;
    int is_subscript = FALSE;
    int is_superscript = FALSE;

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    para_num_string[j++] = ' ';

	    if (current_pgf && current_pgf->pgfnumfont)
	        {
		    struct FMFont *font = current_pgf->pgfnumfont;

		    if (font->weight && !strcmp(font->weight, "Bold"))
			{
			    strcpy(&para_num_string[j], "<paranum><B>");
			    j += 3;
			    is_bold = TRUE;
			}
		    if (font->angle && (!strcmp(font->angle, "Italic") ||
					!strcmp(font->angle, "Oblique")))
			{
			    strcpy(&para_num_string[j], "<I>");
			    j += 3;
			    is_italic = TRUE;
			}
		    if (font->underline)
			{
			    strcpy(&para_num_string[j], "<U>");
			    j += 3;
			    is_underline = TRUE;
			}
		    if (font->subscript)
			{
			    strcpy(&para_num_string[j], "<EM SUB>");
			    j += 8;
			    is_subscript = TRUE;
			}
		    if (font->superscript)
			{
			    strcpy(&para_num_string[j], "<EM SUP>");
			    j += 8;
			    is_superscript = TRUE;
			}
		}

	    i++;

	    while (yytext[i] && yytext[i] != 39)
		{
		    if (yytext[i] == 92)			
			{
			    if (yytext[i+1] == 't')
				{	
				    para_num_string[j++] = ' ';
				    i++;
				}
			    else
				para_num_string[j++] = yytext[i];
			}
		    else
			para_num_string[j++] = yytext[i];
		    i++;
		}

	    if (is_superscript)
		{
		    strcpy(&para_num_string[j], "</EM>");
		    j += 5;
		}
	    if (is_subscript)
		{
		    strcpy(&para_num_string[j], "</EM>");
		    j += 5;
		}
	    if (is_underline)
		{
		    strcpy(&para_num_string[j], "</U>");
		    j += 4;
		}
	    if (is_italic)
		{
		    strcpy(&para_num_string[j], "</I>");
		    j += 4;
		}
	    if (is_bold)
		{
		    strcpy(&para_num_string[j], "</B>");
		    j += 4;
		}

	    para_num_string[j++] = ' ';
	    para_num_string[j] = 0;

	    if (!strcmp(para_num_string, "- "))
		strcpy(para_num_string, "<P><LI>");
	}			
    else
	{
	    para_num_string[0] = 0;
	}
}


/******************************************************************************
  void GetReferenceFile() :
        Input  : Uses yytext
	Output : Reference file
 *****************************************************************************/

void GetReferenceFile()
{
    int i = 0, j = 0;
    char *ptr;

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    i++;
	    while (yytext[i] && yytext[i] != 39)
		{
		    if (yytext[i] == '<')			
			{
			    if (yytext[i+1] == 'c')
				{	
				    i++;		/* Get by <c\> */
				    i++;
				    i++;
				    i++;
				    while (yytext[i] && yytext[i] != '<' && yytext[i] != 39)
					{
					    reference_filename[j++] = yytext[i++];
					}
				    reference_filename[j]=0;
				    if ((ptr = (char *) strstr(reference_filename, ".mif")) != NULL)
					{
					    *ptr = 0;
					}
				    if (STRIP && ((ptr = (char *) strstr(reference_filename, ".doc")) != NULL))
					{
					    *ptr = 0;
					}
				    /* if (!strcmp(TRANSFilename, reference_filename))
				           reference_filename[0] = 0;
				       else	*/
				    strcat(reference_filename, ".html");
				    return;
				}
			}
		    i++;
		}

	    /* Jump inside file */
	    strcpy(reference_filename, TRANSBasename);
	    strcat(reference_filename, ".html");
	}			
    else 
	{
	    /* Jump inside file */
	    strcpy(reference_filename, TRANSBasename);
	    strcat(reference_filename, ".html");
	}
}

	    

/******************************************************************************
  ParaStart() :
 *****************************************************************************/

void ParaStart()
{
}


/******************************************************************************
  PgfStart() :
 *****************************************************************************/

void PgfStart()
{
    if (in_pgf_catalog)
	{
	    struct FMPgf *new_pgf =
	        (struct FMPgf *) calloc(1, sizeof(struct FMPgf));

	    new_pgf->font = (struct FMFont *) calloc(1, sizeof(struct FMFont));

	    new_pgf->tag = NULL;
	    new_pgf->alignment = 'L';
	    new_pgf->topseparator = FALSE;
	    new_pgf->botseparator = FALSE;
	    new_pgf->font->tag = NULL;
	    new_pgf->font->weight = NULL;
	    new_pgf->font->angle = NULL;
	    new_pgf->font->underline = FALSE;
	    new_pgf->font->subscript = FALSE;
	    new_pgf->font->superscript = FALSE;
	    new_pgf->font->next_font = NULL;
	    new_pgf->pgfnumfontstr = NULL;
	    new_pgf->pgfnumfont = NULL;
	    new_pgf->next_pgf = NULL;

	    if (!pgf_list)
	        pgf_list = new_pgf;
	    else
	        current_pgf->next_pgf = new_pgf;

	    current_pgf = new_pgf;
	}
}


/******************************************************************************
  PgfEnd() : 
 *****************************************************************************/

void PgfEnd()
{
}


/******************************************************************************
  ParaEnd() : If the paragraph contained text, then output the end
		   paragraph tag. 
 *****************************************************************************/

void ParaEnd()
{
    int i;
    
    /* djf: Have to reset font in table title, or overlapping HTML tags 
	    will confuse Mosaic causing <PRE> not to have any effect on
	    table.  Inside body of table fonts are set on a cell-by-cell
	    basis so reset is appropriate.  In table header and footer
	    fonts continue until explicitly reset, just as for normal
	    textflows (or at least that seems to be the rule). */
    if (in_table_title || in_table_body)
        ResetFont();

    if (in_string || para_num_string[0])
	{
	    in_string = FALSE;
	    PrintHeaderTypeEnd();
	}

    for (i = 0; i < tabfig_count; i++)
        {
	    if (tabfig[i] == FIGURE)
	        InsertFigure(f, tabfig_ids[i]);
	    else
	        InsertTable(f, tabfig_ids[i]);
	}	
    tabfig_count = 0;

    /* Ensure that old paragraph numbering is not used */
    para_num_font[0] = 0;
    para_num_string[0] = 0;
}


/******************************************************************************
  FontStart() :
 *****************************************************************************/

void FontStart()
{
    in_font = TRUE;
    font_openers[0] = 0;

    if (in_font_catalog)
	{
	    struct FMFont *new_font =
	      (struct FMFont *) calloc(1, sizeof(struct FMFont));

	    new_font->tag = NULL;
	    new_font->weight = NULL;
	    new_font->angle = NULL;
	    new_font->underline = FALSE;
	    new_font->subscript = FALSE;
	    new_font->superscript = FALSE;
	    new_font->next_font = NULL;

	    if (!font_list)
	        font_list = new_font;
	    else
	        current_font->next_font = new_font;

	    current_font = new_font;
	}
}


/******************************************************************************
  FontEnd() : 
 *****************************************************************************/

void FontEnd()
{
    in_font = FALSE;

    if (!in_font_catalog || !current_font->tag || !*current_font->tag)
        {
	    strcat(paragraph, font_openers);
	    return;
	}

    for (current_pgf = pgf_list; current_pgf != NULL;
	 current_pgf = current_pgf->next_pgf)
        if (current_pgf->pgfnumfontstr &&
	    !strcmp(current_pgf->pgfnumfontstr, current_font->tag))
	    current_pgf->pgfnumfont = current_font;
}


/******************************************************************************
  TextFlowStart() : Change output file
 *****************************************************************************/

void TextFlowStart()
{
    debug("TextFlowStart\n");

    in_textflow = TRUE;
    f = flowfile;
}


/******************************************************************************
  int GetIDNextLine() :
        Input  : Uses yytext
	Output : 
 *****************************************************************************/

int GetIDNextLine()
{
    int i = 0;

    while(yytext[i] && yytext[i] != 'D')
        i++;
    i++;

    return atoi(&(yytext[i]));
}


/******************************************************************************
  PrintTextRectID() :
        Input  : Uses yytext
	Output : <<ID>>, where ID is the ID of the textrect
 *****************************************************************************/

void PrintTextRectID()
{
    current_textrect_id = GetIDNextLine();

    if (header == FIGURE && frame_ids[current_textrect_id])
	fprintf(f, "\n>>%d<<%d>>", frame_ids[current_textrect_id],
		current_textrect_id);
    else if (in_body_page || in_textflow)
	fprintf(f, "\n<<%d>>", current_textrect_id);
}


/******************************************************************************
  void ResetFont() : Reset Font settings.
 *****************************************************************************/

void ResetFont()
{
    if (font_bold)
	{
	    strcat(paragraph, "</B>");
	    font_bold = FALSE;
	}
    if (font_italic)
	{
	    strcat(paragraph, "</I>");
	    font_italic = FALSE;
	}
    if (font_underline)
	{
	    strcat(paragraph, "</U>");
	    font_underline = FALSE;
	}
    if (font_subscript)
	{
	    strcat(paragraph, "</EM>");
	    font_subscript = FALSE;
	}
    if (font_superscript)
	{
	    strcat(paragraph, "</EM>");
	    font_superscript = FALSE;
	}
    if (in_code)
        {
	    strcat(paragraph, "</CODE>");
	    in_code = FALSE;
	}
    font_type = NORMAL;
}



/******************************************************************************
  FontFamily() :
 *****************************************************************************/

void FontFamily()
{
    char font_name[50];
    int i = 10;			/* Skip "<FFamily `" */
    int j = 0;

    if (in_pgf_catalog || in_font_catalog)
        return;

    while (yytext[i] && yytext[i] != 39)
	{
	    font_name[j++] = yytext[i];
	    i++;
	}

    font_name[j] = 0;

    if (in_code &&
	(strcasecmp(font_name, "Courier") || !strstr(font_name, "Mono")))
        {
            strcat(paragraph, "</CODE>");
	    in_code = FALSE;
	}

    if (yytext[i])
	{
	    if (!strcasecmp(font_name, "Symbol"))
		{
		    font_type = SYMBOL;
		}
	    else if (!strcasecmp(font_name, "ZapfDingbats"))
		{
		    font_type = DINGBATS;
		}
	    else if (!strcasecmp(font_name, "Courier") ||
		     (strstr(font_name, "Mono")))
		{
		    if (font_type != COURIER && !in_program)
			{
			    font_type = COURIER;
			    if (in_font)
			        strcat(font_openers, "<CODE>");
			    else
			        strcat(paragraph, "<CODE>");
			    in_code = TRUE;
			}
		}
	    else 
		{
		    font_type = NORMAL;
		}
	}
}


/******************************************************************************
  InsertSpecialChar() :
 *****************************************************************************/

void InsertSpecialChar()
{
    char char_name[50];
    int i = 6;			/* Skip "<Char " */
    int j = 0;

    if (!in_string)
	PrintHeaderTypeStart();
    in_string = TRUE;
	    
    while (yytext[i] && yytext[i] != 32)
	{
	    char_name[j++] = yytext[i];
	    i++;
	}
    char_name[j] = 0;

    if (font_type == SYMBOL)
	{
	    if (yytext[i])
		{
		    if (!strcmp(char_name, "EmDash"))
			strcat(paragraph, kEM_DASH_PREF);
		    else if (!strcmp(char_name, "EnDash"))
			strcat(paragraph, kEN_DASH_PREF);
		    else if (!strcmp(char_name, "EmSpace"))
			strcat(paragraph, "   ");
		    else if (!strcmp(char_name, "EnSpace"))
			strcat(paragraph, "  ");
		    else if (!strcmp(char_name, "HardSpace"))
			strcat(paragraph, " ");
		    else if (!strcmp(char_name, "Pound"))
			strcat(paragraph, "\\xa3\\");
		    else if (!strcmp(char_name, "Cent"))
			strcat(paragraph, "\\xa2\\");
		    else if (!strcmp(char_name, "Yen"))
			strcat(paragraph, "\\xb4\\");
		    else if (!strcmp(char_name, "HardReturn"))
		        {
			    /* ejr: long lists sometimes are broken with
			     *      a hard return instead of a tab.
			     */
			    if (header >= GLOSSARY1 && header <= COMGLOSSARY3
				|| header == ACRONYM)
			        strcat(paragraph, "<DD>");
			    else if (header == PROGRAM)
			        strcat(paragraph, "\n");
			    else
			        strcat(paragraph, "\n<BR>");
			}
		    else if (!strcmp(char_name, "Tab"))
			{
			    if (header >= GLOSSARY1 && header <= COMGLOSSARY3
				|| header == ACRONYM)
				strcat(paragraph, "<DD>");
			    else
				strcat(paragraph, "  ");
			}
		    else if (in_tables && (!strcmp(char_name, "SoftHyphen") ||
					   !strcmp(char_name, "HardHyphen")))
			strcat(paragraph, "-");
		    else if (!(strcmp(char_name, "SoftHyphen") && 
			     strcmp(char_name, "HardHyphen") &&
			     strcmp(char_name, "DiscHyphen") && 
			     strcmp(char_name, "NoHyphen")))
			{
			    /* fprintf(stderr, "trans: Special SYMBOL character not handled: \"%s\"\n", char_name); */
			}
		}
	}
    else if (font_type == DINGBATS)
	{
	    if (!QUIET)
	        fprintf(stderr,
			"trans: Special Dingbat char not handled: \"%s\"\n",
			char_name);
	}
    else
	{
	    if (yytext[i])
		{
		    if (!strcmp(char_name, "EmDash"))
			strcat(paragraph,  kEM_DASH_PREF);
		    else if (!strcmp(char_name, "EnDash"))
			strcat(paragraph, kEN_DASH_PREF);
		    else if (!strcmp(char_name, "EmSpace"))
			strcat(paragraph, "   ");
		    else if (!strcmp(char_name, "EnSpace"))
			strcat(paragraph, "  ");
		    else if (!strcmp(char_name, "HardSpace"))
			strcat(paragraph, " ");
		    else if (!strcmp(char_name, "Pound"))
			strcat(paragraph, "\xa3");
		    else if (!strcmp(char_name, "Cent"))
			strcat(paragraph, "\xa2");
		    else if (!strcmp(char_name, "Yen"))
			strcat(paragraph, "\xa5");
		    else if (!strcmp(char_name, "HardReturn"))
			{
			    /* ejr: long lists sometimes are broken with
			     *      a hard return instead of a tab.
			     */
			    if (header >= GLOSSARY1 &&
				header <= COMGLOSSARY3 || header == ACRONYM)
			        strcat(paragraph, "<DD>");
			    else if (header == PROGRAM)
			        strcat(paragraph, "\n");
                            else if (!in_tables)
                                strcat(paragraph, "\n<BR>");
			}
		    else if (!strcmp(char_name, "Tab"))
			{
			    if (header >= GLOSSARY1 && 
				header <= COMGLOSSARY3 || header == ACRONYM)
				strcat(paragraph, "<DD>");
			    else
				strcat(paragraph, "  ");
			}
		    else if (!(strcmp(char_name, "SoftHyphen") && strcmp(char_name, "HardHyphen") &&
			     strcmp(char_name, "DiscHyphen") && strcmp(char_name, "NoHyphen")))
			{
			    /* fprintf(stderr, "trans: Special character not handled: \"%s\"\n", char_name); */
			}
		}
	}
}


/******************************************************************************
  TextRectStart() :
 *****************************************************************************/

void TextRectStart()
{
    debug("TextRectStart\n");

    PrintTextRectID();

    if (in_frame || in_inline_frame)
	{
	    textrect_ids[textrect_count] = current_textrect_id;
	    frame_ids[current_textrect_id] = current_frame_id;
	    textrect_count++;
	}
}


/******************************************************************************
  TextRectEnd() :
 *****************************************************************************/

void TextRectEnd()
{
    debug("TextRectEnd\n");
}


/******************************************************************************
  InsertFigure() :
 *****************************************************************************/

void InsertFigure(f, id)
      FILE *f;
      int id;
{
    sprintf(frame_filename, "%s.fig_%d", TRANSBasename, id);
    fprintf(f, "<P><A HREF=\"%s.ps\"><IMG SRC=\"%s.gif\"></A>\n", 
	    frame_filename, frame_filename);
}


/******************************************************************************
  InsertInlineFigure() :
 *****************************************************************************/

void InsertInlineFigure(f, id)
      FILE *f;
      int id;
{
    char frame_link[1000];
    if (!in_string)
	PrintHeaderTypeStart();
    in_string = TRUE;
	    
    if (in_tables)
	{
	    sprintf(frame_link, "<A HREF = \"%s.fig_%d.gif\">(fig)</A>", 
		    TRANSBasename, id);
	}
    else
	{
	    sprintf(frame_link,
		    "<A HREF = \"%s.fig_%d.ps\"><IMG SRC=\"%s.fig_%d.gif\"></A>", 
		    TRANSBasename, id, TRANSBasename, id);
	}
    strcat(paragraph, frame_link);
}


/******************************************************************************
  SetFrameType() :
 *****************************************************************************/

void SetFrameType()
{
    if (NOGRAPHICS)
        return;

    frame_id_type[frame_count][0] = current_frame_id;

    if (strstr(yytext, "Inline"))
	frame_id_type[frame_count][1] = INLINEFRAME;
    else
	frame_id_type[frame_count][1] = OTHERFRAME;
}


/******************************************************************************
  GetFrameType() :
 *****************************************************************************/

int GetFrameType(frame_id)
      int frame_id;
{
    int i;
    for (i = 0; i < frame_count; i++)
        if (frame_id_type[i][0] == frame_id)
	    return frame_id_type[i][1];
    return -1;
}


/******************************************************************************
  ParagraphAlignment() :
        Input  : Uses yytext
	Output : Set flag in paragraph struct
 *****************************************************************************/

void ParagraphAlignment()
{
    if (in_pgf_catalog)
        {
	    if (strstr(yytext, "Left")) /* include LeftRight, i.e. justified */
	        current_pgf->alignment = 'L';
	    else if (strstr(yytext, "Center"))
	        current_pgf->alignment = 'C';
	    else
	        current_pgf->alignment = 'R';
	}
    else if (current_cell)
        {
	    if (strstr(yytext, "Left")) /* include LeftRight, i.e. justified */
	        current_cell->alignment = 'L';
	    else if (strstr(yytext, "Center"))
	        current_cell->alignment = 'C';
	    else
	        current_cell->alignment = 'R';
	}
      
}


/******************************************************************************
  ParagraphTopSeparator() :
        Input  : Uses yytext
	Output : Set flag in paragraph struct
 *****************************************************************************/

void ParagraphTopSeparator()
{
    char *value = GetValue();

    if (value)
        {
	    if (in_pgf_catalog)
	        current_pgf->topseparator = TRUE;
	    else if (!in_font_catalog)
	        fprintf(f,"<HR>\n");
	    free(value);
	}
}


/******************************************************************************
  ParagraphBottomSeparator() :
        Input  : Uses yytext
	Output : Set flag in paragraph struct
 *****************************************************************************/

void ParagraphBottomSeparator()
{
    char *value = GetValue();

    if (value)
        {
	    if (in_pgf_catalog)
	        current_pgf->botseparator = TRUE;
	    else if (!in_font_catalog)
	        separator_pending = TRUE;
	    free(value);
	}
}


/******************************************************************************
  FontWeight() :
        Input  : Uses yytext
 *****************************************************************************/

void FontWeight()
{
    if (in_pgf_catalog)
        current_pgf->font->weight = GetValue();
    else if (in_font_catalog)
        current_font->weight = GetValue();
    else
	{
	    char *value = GetValue();
	    ChangeFontWeight(value);
	    free(value);
	}
}


/******************************************************************************
  ChangeFontWeight() :
        Input  : Value string, either stored or derived from current
		 value of yytext
 *****************************************************************************/

void ChangeFontWeight(value)
      char *value;
{
    if (!value)
        return;

    /* djf: Mosaic does strange things if you insert formatting tags inside
            headings.  I've disabled bolding in headings until such time as
	    Mosaic starts behaving sensibly. */
    if (in_heading)
        return;

    if (!strcmp(value, "Bold"))
	{
	    if (!font_bold)
	        {
		    if (in_font)
		        strcat(font_openers, "<B>");
		    else
		        strcat(paragraph, "<B>");
		    font_bold = TRUE;
	        }
	}
    else if (font_bold)
	{
	    strcat(paragraph, "</B>");
	    font_bold = FALSE;
	}
}


/******************************************************************************
  FontAngle() :
        Input  : Uses yytext
 *****************************************************************************/

void FontAngle()
{
    if (in_pgf_catalog)
        current_pgf->font->angle = GetValue();
    else if (in_font_catalog)
        current_font->angle = GetValue();
    else
	{
	    char *value = GetValue();
	    ChangeFontAngle(value);
	    free(value);
	}
}


/******************************************************************************
  ChangeFontAngle() :
        Input  : Value string, either stored or derived from current
		 value of yytext
 *****************************************************************************/

void ChangeFontAngle(value)
    char *value;
{
    if (!value)
        return;

    /* djf: Mosaic does strange things if you insert formatting tags inside
            headings.  I've disabled italics in headings until such time as
	    Mosaic starts behaving sensibly. */
    if (in_heading)
        return;

    /* afeb: added Oblique */
    if (!strcmp(value, "Italic") || !strcmp(value, "Oblique"))
	{
	    if (!font_italic)
	        {
		    strcat(paragraph, "<I>");
		    font_italic = TRUE;
		}
	}
    else if (font_italic)
	{
	    strcat(paragraph, "</I>");
	    font_italic = FALSE;
	}
}


/******************************************************************************
  FontUnderline() :
        Input  : Uses yytext
 *****************************************************************************/

void FontUnderline()
{
    int is_underline = strstr(yytext, "Yes") != NULL;
    if (in_pgf_catalog)
        current_pgf->font->underline = is_underline;
    else if (in_font_catalog)
        current_font->underline = is_underline;
    else
        ChangeFontUnderline(is_underline);
}


/******************************************************************************
  ChangeFontUnderline() :
        Input  : Uses current or stored value of yytext
 *****************************************************************************/

void ChangeFontUnderline(is_underline)
    int is_underline;
{
    /* djf: Mosaic does strange things if you insert formatting tags inside
            headings.  I've disabled underlining in headings until such time as
	    Mosaic starts behaving sensibly. */
    if (in_heading)
        return;

    if (is_underline)
	{
	    if (!font_underline)
	        {
		    if (in_font)
		        strcat(font_openers, "<U>");
		    else
		        strcat(paragraph, "<U>");
		    font_underline = TRUE;
		}
	}
    else if (font_underline)
	{
	    strcat(paragraph, "</U>");
	    font_underline = FALSE;
	}
}


/******************************************************************************
  FontSubscript() :
        Input  : Uses yytext
 *****************************************************************************/

void FontSubscript()
{
    int is_subscript = strstr(yytext, "Yes") != NULL;
    if (in_pgf_catalog)
        current_pgf->font->subscript = is_subscript;
    else if (in_font_catalog)
        current_font->subscript = is_subscript;
    else
        ChangeFontSubscript(is_subscript);
}


/******************************************************************************
  ChangeFontSubscript() :
        Input  : Uses current or stored value of yytext
 *****************************************************************************/

void ChangeFontSubscript(is_subscript)
      int is_subscript;
{
    /* djf: Mosaic does strange things if you insert formatting tags inside
            headings.  I've disabled subscripting in headings until such as
	    time Mosaic starts behaving sensibly. */
    if (in_heading)
        return;

    if (is_subscript)
	{
	    if (!font_subscript)
	        {
		    if (in_font)
		        strcat(font_openers, "<EM SUB>");
		    else
		        strcat(paragraph, "<EM SUB>");
		    font_subscript = TRUE;
		}
	}
    else if (font_subscript)
	{
	    strcat(paragraph, "</EM>");
	    font_subscript = FALSE;
	}
}


/******************************************************************************
  FontSuperscript() :
        Input  : Uses yytext
 *****************************************************************************/

void FontSuperscript()
{
    int is_superscript = strstr(yytext, "Yes") != NULL;
    if (in_pgf_catalog)
        current_pgf->font->superscript = is_superscript;
    else if (in_font_catalog)
        current_font->superscript = is_superscript;
    else
        ChangeFontSuperscript(is_superscript);
}


/******************************************************************************
  ChangeFontSuperscript() :
        Input  : Uses current or stored value of yytext
 *****************************************************************************/

void ChangeFontSuperscript(is_superscript)
      int is_superscript;
{
    /* djf: Mosaic does strange things if you insert formatting tags inside
            headings.  I've disabled superscripting in headings until such as
	    time Mosaic starts behaving sensibly. */
    if (in_heading)
        return;

    if (is_superscript)
	{
	    if (!font_superscript)
	        {
		    if (in_font)
		        strcat(font_openers, "<EM SUP>");
		    else
		        strcat(paragraph, "<EM SUP>");
		    font_superscript = TRUE;
		}
	}
    else if (font_superscript)
	{
	    strcat(paragraph, "</EM >");
	    font_superscript = FALSE;
	}
}


/******************************************************************************
  TextFlowTag() :
        Input  : Uses yytext
	Output : Text flow tag string
 *****************************************************************************/

void TextFlowTag()
{
    int i = 0, j = 0;

    while(yytext[i] && yytext[i] != 96)
        i++;

    if (yytext[i] == 96)
	{
	    i++;
	    while (yytext[i] && yytext[i] != 39)
	        text_flow_tag[j++] = yytext[i++];
	    text_flow_tag[j] = 0;
	}			
    else
        text_flow_tag[0] = 0;
}



/******************************************************************************
 void TextFlowEnd() : 
 *****************************************************************************/

void TextFlowEnd()
{
    debug("TextFlowEnd\n");

    in_textflow = FALSE;

    if (depth)
        Ascend(depth);

    if (in_reference)
	{
	    fprintf(f, "</DL>\n");
	    in_reference = FALSE;
	}
    else if (in_glossary)
	{
	    switch (in_glossary)
		{
		  case GLOSSARY1:
		  case COMGLOSSARY1:
		    fprintf(f, "</DL>\n");
		    break;
		  case GLOSSARY2:
		  case COMGLOSSARY2:
		    fprintf(f, "</DL></UL>\n");
		    break;
		  case GLOSSARY3:
		  case COMGLOSSARY3:
		    fprintf(f, "</DL></UL></UL>\n");
		    break;
		    
		  default:
		    if (!QUIET)
		      fprintf(stderr, "trans: Don't know glossary state %d\n",
			      in_glossary);
		}
	    in_glossary = FALSE;
	    
	}
    else if (in_program)
	{
	    fprintf(f, "</PRE>\n");
	    in_program = FALSE;
	}
    else if (in_acronym)
	{
	    fprintf(f, "</DL>\n");
	    in_acronym = FALSE;
	}

    text_flow_tag[0] = 0;
}



/******************************************************************************
  void InlineFrameStart() : 
        Input  : 
	Output : 
 *****************************************************************************/

void InlineFrameStart()
{
    debug("InlineFrameStart\n");

    if (NOGRAPHICS || !in_body_page)
        return;

    in_inline_frame = TRUE;
    textrect_count = 0;
    current_frame_id = inlineframe_num++;

    sprintf(frame_filename, "%s.fig_I%d.mif", TRANSFilename, current_frame_id);
    sprintf(sys_call, "cp %s %s", HEADFilename, frame_filename);
    if (!QUIET)
        fprintf(stderr, "trans: Making FrameMaker MIF file \"%s\"\n",
		frame_filename);
    system(sys_call);

    framef = fopen(frame_filename, "a");
    if (!framef)
	{
	    fprintf(stderr, "trans: ERROR: Could not open file \"%s\"\n",
		    frame_filename);
	    exit(-1);
	}
    fprintf(framef, "%s\n", yytext);
}


/******************************************************************************
  PrintFrameItem() :
 *****************************************************************************/

void PrintFrameItem()
{
    if (NOGRAPHICS)
	return;

    if (in_inline_frame)
        fprintf(framef, "%s\n", yytext);
}


/******************************************************************************
  void InlineFrameEnd() :
        Input  : 
	Output :
 *****************************************************************************/

void InlineFrameEnd()
{
    int i;
    struct stat stbuf;

    debug("InlineFrameEnd\n");

    if (NOGRAPHICS || !in_body_page)
        return;

    in_inline_frame = FALSE;
    fprintf(framef, "%s\n", yytext);
    fclose(framef);

    if (textrect_count > 0)
	{
	    sprintf(sys_call, "-SCRDIR-/addtextrect %s %s", TRANSFilename, frame_filename);

	    for (i = 0; i < textrect_count; i++) 
		{
		    char temp_num[10];
		    strcat(sys_call, " ");
		    sprintf(temp_num, " %d%c", textrect_ids[i], 0);
		    strcat(sys_call, temp_num);
		}
	    system(sys_call);
	}

    framef = fopen(frame_filename, "a");
    if (!framef)
	{
	    fprintf(stderr, "trans: ERROR: Could not open file \"%s\"\n", frame_filename);
	    exit(-1);
	}
    fprintf(framef, "# End of MIFFile\n");
    fclose(framef);
    
    /* Translate MIF figure to GIF */
    sprintf(sys_call, "-SCRDIR-/mif2gif %s %s %s %s %s %s", NOCROP ? "" : "-c",
	    XDPI ? XDPI : "", YDPI ? YDPI : "", QUIET ? "-q" : "",
	    DEBUG ? "" : "-r", frame_filename);
    system(sys_call);
	    
    /* Remove GIF if empty (has length 35 or 0) */
    frame_filename[strlen(frame_filename) - 3] = 'g';
    if (stat(frame_filename, &stbuf) == 0)
	if (stbuf.st_size == 35 || stbuf.st_size == 0) 
	    {
	        if (!QUIET)
		    fprintf(stderr, "trans: Deleting empty GIF file: %s\n",
			    frame_filename);
		unlink(frame_filename);
	    }

    fprintf(f, "\n<A HREF = \"%s\"><IMG SRC=\"%s\"></A>\n", 
	    frame_filename, frame_filename);
}


/******************************************************************************
  void StripSpace() :
 *****************************************************************************/

void StripSpace(txt)
      char *txt;
{
    int length = strlen(txt);
    int i;
    char *txtp;
    
    txtp = txt + length - 1;

    for (i = length - 1; i >= 0; i--, txtp--)
	{
	    if (*txtp == 32)
		*txtp = 0;
	    else
		return;
	}
}


/******************************************************************************
  void GetTextOnly :
 *****************************************************************************/

char *GetTextOnly(from)
      char *from;
{
    char *fromp = from;
    char *to = calloc(strlen(from) + 1, 1);
    char *top = to;

    while (*fromp)
	{
	    if (*fromp == '<')
		{
		    while (*fromp && *fromp != '>')
		        fromp++;
		    if (!*fromp)
		        return to;
		}
	    else
		{
		    *top = *fromp;
		    top++;
		}
	    fromp++;
	}

    return to;
}
	

/******************************************************************************
  void HandleParaLine()
 *****************************************************************************/

void HandleParaLine()
{
    if (in_tables && !in_footnote) 
        NewCellLine();

    if (header == PROGRAM && in_string) 
	strcat(paragraph, "\n");
}


/******************************************************************************
 void main: added by popular demand
 *****************************************************************************/

void main()
{
    yyparse();
}


/******************************************************************************
 Variable handling
 *****************************************************************************/

#define MAX_VARIABLES 512

typedef struct sVarLookup
{
    char *fVarName;
    char *fVarDef;
}
TVarLookup;

TVarLookup varLookup[MAX_VARIABLES] = {NULL,};
int numVariables = 0;
char *lastVariable = NULL;


/******************************************************************************
  char *GetVarDef() :
	Input  : Uses yytext
	Output : Text in yytext between ` '
 *****************************************************************************/

char *GetVarDef()
{
    int i = 0, j = 0;
    char *quoted = malloc(strlen(yytext) * 5);  /* Room enough for < > and & */
    
    while(yytext[i] && yytext[i] != 96)
        i++;
    
    if (yytext[i] == 96)
	{
	    i++;
	    while(yytext[i] && yytext[i] != 39)
		{
		    if (yytext[i] == '<')
			{
			    while(yytext[i] && !(yytext[i] == '\\' && 
						 yytext[i+1] == '>'))
				i++;
			    if (yytext[i] && yytext[i] == '\\')
				i++;
			}
		    else if (yytext[i] == '\\' && yytext[i+1] == 'q')
			{
			    quoted[j++] = '\'';
			    i++;
			}
		    else if (yytext[i] == '\\' && yytext[i+1] == 'x')
			{
			    if (yytext[i+2] == 'a' && yytext[i+3] == 'a')
				{
				    quoted[j++] = '(';
				    quoted[j++] = 't';
				    quoted[j++] = 'm';
				    quoted[j++] = ')';
				}
			    i++;
			    i++;
			    i++;
			}
		    else if (yytext[i] == '\\' && yytext[i+1] == '\\')
			{
			    quoted[j++] = '\\';
			    i++;
			}
		    else if (yytext[i] == '&')
			{
			    quoted[j++] = '&';
			    quoted[j++] = 'a';
			    quoted[j++] = 'm';
			    quoted[j++] = 'p';
			    quoted[j++] = ';';
			}
		    else
			quoted[j++] = yytext[i];
		    i++;
		}
	    if (yytext[i] == 39)
		{
		    quoted[j] = 0;
		    return quoted;
		}
	    else
		{
		    free(quoted);
		    return NULL;
		}
	}
    else
	{
	    free(quoted);
	    return NULL;
	}
}


/******************************************************************************
  SaveVariable() : Save a variable name
 *****************************************************************************/

SaveVariable()
{
    char *theVariable;
    
    theVariable = GetContents();
    lastVariable = theVariable;
}


/******************************************************************************
  SaveVariableDef() : Save the variable's definition (follows name)
 *****************************************************************************/

SaveVariableDef()
{
    char *theDef;
    
    theDef = GetVarDef();
    
    if (numVariables < MAX_VARIABLES)
	{
	    varLookup[numVariables].fVarName = lastVariable;
	    varLookup[numVariables].fVarDef  = theDef;
	    numVariables++;
	}
}


/******************************************************************************
  NoteVariName() : Replaces variable name with its definition
 *****************************************************************************/

NoteVariName()
{
    int loop;
    char *theVariable = GetContents();
    
    for (loop = 0; loop < numVariables; loop++)
	if (strcmp(theVariable, varLookup[loop].fVarName) == 0)
	    strcat(paragraph, varLookup[loop].fVarDef);
}

                                     
